{% extends "base.html" %}
{% block title %}
All Files &ndash; {{ project }}
{% endblock %}
{% block body %}
      <div class="row" id='text'>
        <div class="col-lg-12">
			 <h1>Source Files</h1>
			 <table class="table table-striped">
			 <thead><tr><th>File</th><th>Description</th></tr></thead>
			 <tbody>
			 {% for src in project.allfiles|sort(attribute='name') %}
			   <tr><td>{{ src }}</td><td>{{ src.meta['summary'] }}</td></tr>
			 {% endfor %}
			 </tbody></table>
			 {{ project.filegraph }}
        </div>
      </div>
{% endblock %}
